/*
 * Copyright notice from original mutt:
 * Copyright (C) 1998 Michael R. Elkins <me@mutt.org>
 * Copyright (C) 1999-2005 Brendan Cully <brendan@kublai.com>
 * Copyright (C) 1999-2000 Tommi Komulainen <Tommi.Komulainen@iki.fi>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/lib-lib.h>

#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>

#include <lib-ui/lib-ui.h>

#include "mutt.h"
#include "globals.h"

#include "unix.h"
#include "mutt_socket.h"
#include "mutt_tunnel.h"
#include "mutt_signal.h"

#ifdef HAVE_LIBIDN
#include <idna.h>
#endif
#include "mutt_idna.h"

/* support for multiple socket connections */
static CONNECTION *Connections = NULL;

/* Wrappers */
int mutt_socket_open (CONNECTION * conn)
{
  int rc;

  if (m_strlen(Preconnect)) {
    rc = mutt_system (Preconnect);
    if (rc) {
      mutt_perror (_("Preconnect command failed."));
      mutt_sleep (1);
      return -1;
    }
  }

  return conn->conn_open(conn);
}

int mutt_socket_close (CONNECTION * conn)
{
  int rc = -1;

  if (conn->fd >= 0)
    rc = conn->conn_close (conn);

  conn->fd = -1;
  conn->ssf = 0;

  return rc;
}

int mutt_socket_read (CONNECTION * conn, char *buf, ssize_t len)
{
  int rc;

  if (conn->fd < 0) {
    return -1;
  }

  rc = conn->conn_read (conn, buf, len);
  /* EOF */
  if (rc == 0) {
    mutt_error (_("Connection to %s closed"), conn->account.host);
    mutt_sleep (2);
  }
  if (rc <= 0)
    mutt_socket_close (conn);

  return rc;
}

int mutt_socket_write(CONNECTION * conn, const char *buf)
{
  int rc;
  int len;

  if (conn->fd < 0) {
    return -1;
  }

  len = m_strlen(buf);
  if ((rc = conn->conn_write (conn, buf, len)) < 0) {
    mutt_socket_close (conn);

    return -1;
  }

  return rc;
}

/* simple read buffering to speed things up. */
int mutt_socket_readchar (CONNECTION * conn, char *c)
{
  if (conn->bufpos >= conn->available) {
    if (conn->fd >= 0)
      conn->available =
        conn->conn_read (conn, conn->inbuf, sizeof (conn->inbuf));
    else {
      return -1;
    }
    conn->bufpos = 0;
    if (conn->available == 0) {
      mutt_error (_("Connection to %s closed"), conn->account.host);
      mutt_sleep (2);
    }
    if (conn->available <= 0) {
      mutt_socket_close (conn);
      return -1;
    }
  }
  *c = conn->inbuf[conn->bufpos];
  conn->bufpos++;
  return 1;
}

int mutt_socket_readln(char *buf, ssize_t buflen, CONNECTION * conn)
{
  char ch;
  ssize_t i;

  for (i = 0; i < buflen - 1; i++) {
    if (mutt_socket_readchar (conn, &ch) != 1) {
      buf[i] = '\0';
      return -1;
    }

    if (ch == '\n')
      break;
    buf[i] = ch;
  }

  /* strip \r from \r\n termination */
  if (i && buf[i - 1] == '\r')
    buf[--i] = '\0';
  else
    buf[i] = '\0';

  /* number of bytes read, not m_strlen*/
  return i + 1;
}

int mutt_socket_readln2(buffer_t *buf, CONNECTION *conn)
{
    char ch;

    while (mutt_socket_readchar(conn, &ch) == 1) {
        if (ch == '\n') {
            if (buf->len && buf->data[buf->len - 1] == '\r') {
                buf->data[--buf->len] = '\0';
            }
            return 0;
        }
        buffer_addch(buf, ch);
    }

    return -1;
}

CONNECTION *mutt_socket_head (void)
{
  return Connections;
}

/* mutt_socket_free: remove connection from connection list and free it */
void mutt_socket_free (CONNECTION * conn)
{
  CONNECTION *iter;
  CONNECTION *tmp;

  iter = Connections;

  /* head is special case, doesn't need prev updated */
  if (iter == conn) {
    Connections = iter->next;
    p_delete(&iter);
    return;
  }

  while (iter->next) {
    if (iter->next == conn) {
      tmp = iter->next;
      iter->next = tmp->next;
      p_delete(&tmp);
      return;
    }
    iter = iter->next;
  }
}

/* mutt_conn_find: find a connection off the list of connections whose
 *   account matches account. If start is not null, only search for
 *   connections after the given connection (allows higher level socket code
 *   to make more fine-grained searches than account info - eg in IMAP we may
 *   wish to find a connection which is not in IMAP_SELECTED state) */
CONNECTION *mutt_conn_find (const CONNECTION * start, const ACCOUNT * account)
{
  CONNECTION *conn;

  conn = start ? start->next : Connections;
  while (conn) {
    if (mutt_account_match (account, &(conn->account)))
      return conn;
    conn = conn->next;
  }

  conn = p_new(CONNECTION, 1);
  conn->fd = -1;
  conn->account = *account;
  conn->next  = Connections;
  Connections = conn;

  if (Tunnel && *Tunnel)
    mutt_tunnel_socket_setup (conn);
  else if (account->has_ssl) {
    if (mutt_ssl_socket_setup (conn) < 0) {
      mutt_socket_free (conn);
      return NULL;
    }
  } else {
    conn->conn_read = raw_socket_read;
    conn->conn_write = raw_socket_write;
    conn->conn_open = raw_socket_open;
    conn->conn_close = raw_socket_close;
  }

  return conn;
}

/* socket_connect: set up to connect to a socket fd. */
static int socket_connect (int fd, struct sockaddr *sa)
{
  int sa_size;
  int save_errno;

  if (sa->sa_family == AF_INET)
    sa_size = sizeof (struct sockaddr_in);
  else if (sa->sa_family == AF_INET6)
    sa_size = sizeof (struct sockaddr_in6);
  else {
    return -1;
  }

  if (ConnectTimeout > 0)
    alarm (ConnectTimeout);

  mutt_allow_interrupt (1);

  save_errno = 0;

  if (connect (fd, sa, sa_size) < 0) {
    save_errno = errno;
    SigInt = 0;                 /* reset in case we caught SIGINTR while in connect() */
  }

  if (ConnectTimeout > 0)
    alarm (0);
  mutt_allow_interrupt (0);

  return save_errno;
}

int raw_socket_close (CONNECTION * conn)
{
  return close (conn->fd);
}

int raw_socket_read (CONNECTION * conn, char *buf, ssize_t len)
{
  int rc;

  if ((rc = read (conn->fd, buf, len)) == -1) {
    mutt_error (_("Error talking to %s (%s)"), conn->account.host,
                strerror (errno));
    mutt_sleep (2);
  }

  return rc;
}

int raw_socket_write (CONNECTION * conn, const char *buf, ssize_t count)
{
  int rc;

  if ((rc = write (conn->fd, buf, count)) == -1) {
    mutt_error (_("Error talking to %s (%s)"), conn->account.host,
                strerror (errno));
    mutt_sleep (2);
  }

  return rc;
}

int raw_socket_open (CONNECTION * conn)
{
  int rc;
  int fd;

  char *host_idna = NULL;

  /* "65536\0" */
  char port[6];
  struct addrinfo hints;
  struct addrinfo *res;
  struct addrinfo *cur;

  /* we accept v4 or v6 STREAM sockets */
  p_clear(&hints, 1);

  hints.ai_family   = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  snprintf (port, sizeof (port), "%d", conn->account.port);

# ifdef HAVE_LIBIDN
  if (idna_to_ascii_lz (conn->account.host, &host_idna, 1) != IDNA_SUCCESS) {
    mutt_error (_("Bad IDN \"%s\"."), conn->account.host);
    return -1;
  }
# else
  host_idna = conn->account.host;
# endif

  mutt_message (_("Looking up %s..."), conn->account.host);

  rc = getaddrinfo (host_idna, port, &hints, &res);

# ifdef HAVE_LIBIDN
  p_delete(&host_idna);
# endif

  if (rc) {
    mutt_error (_("Could not find the host \"%s\""), conn->account.host);
    mutt_sleep (2);
    return -1;
  }

  mutt_message (_("Connecting to %s..."), conn->account.host);

  rc = -1;
  for (cur = res; cur != NULL; cur = cur->ai_next) {
    fd = socket (cur->ai_family, cur->ai_socktype, cur->ai_protocol);
    if (fd >= 0) {
      if ((rc = socket_connect (fd, cur->ai_addr)) == 0) {
        fcntl (fd, F_SETFD, FD_CLOEXEC);
        conn->fd = fd;
        break;
      }
      else
        close (fd);
    }
  }

  freeaddrinfo (res);

  if (rc) {
    mutt_error (_("Could not connect to %s (%s)."), conn->account.host,
                (rc > 0) ? strerror (rc) : _("unknown error"));
    mutt_sleep (2);
    return -1;
  }

  return 0;
}
